/*******************************************************************************
 * Copyright (c) 2008-2011 Supélec.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This file is part of the ModHel'X tool
 * Version 1.1 (April 5, 2008)
 *
 * Contributors:
 *     Computer Science Department, Supélec : initial implementation
 * Contacts:
 *     Cecile.Hardebolle@supelec.fr
 *     Christophe.Jacquet@supelec.fr
 *     Dominique.Marcadet@supelec.fr
 *     Frederic.Boulanger@supelec.fr
 ******************************************************************************/
package fr.supelec.modhelx.engine;


import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Semaphore;
import java.util.logging.FileHandler;
import java.util.logging.Formatter;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;

import fr.supelec.modhelx.core.Model;
import fr.supelec.modhelx.core.time.Clock;
import fr.supelec.modhelx.core.time.InstantValue;
import fr.supelec.modhelx.core.time.TimeFactory;
import fr.supelec.modhelx.engine.instrumentation.Controller;
import fr.supelec.modhelx.engine.instrumentation.TransparentController;

/**
 * 
 * Platform
 */
public class Platform {
	
    static private Model rootModel = null;
    static private Clock rootClock = null;
    
    static private Logger log = Logger.getLogger("fr.supelec.modhelx");
    
    static private Controller controller = new TransparentController(); 
    
    static private InstantValue currentTime = null;
    
    static private int numberOfSnapshots = 0;
    
    static private Semaphore snapshotRequested = new Semaphore(0);
    
    static private long clockUnitInMilliseconds = -1;
    static private long startTimeInMilliseconds = 0;
    static private long dateOfLastSnapshot = -1;

	private static long intervalBetweenSnapshots = -1;
	
	private static List<LagListener> lagListeners = new ArrayList<LagListener>();

    static public void setModel(Model m, Clock clock) {
    	rootModel = m;
    	rootClock = clock;
    }
    
    static public Logger getLog() {
    	return log;
    }
    
    static public void setLogFile(String path) {
    	try {
			log.addHandler(new FileHandler(path));
		} catch (SecurityException e) {
			log.severe(e.getMessage());
		} catch (IOException e) {
			log.severe(e.getMessage());
		}
		log.setUseParentHandlers(false);
    }
    
    static public void setLogLevel(Level newLevel) {
    	log.setLevel(newLevel);
    }
    
    static public void setLogFormatter(Formatter f) {
		Logger l = log;
		while ((l != null) && (l.getHandlers().length == 0)) {
			l = l.getParent();
		}
    	if (l != null) {
    		l.getHandlers()[0].setFormatter(f);
    	}
    }
    
    static public Controller getController() {
    	return controller;
    }
    
    static public void setController(Controller controller) {
    	Platform.controller = controller;
    }
 
    /**
     * The platform will stop to make snapshots either when no
     * model ask for a new one, or when the number specified via
     * this method is reached (0 means don't stop for this reason)  
     */
	public static void setNumberOfSnapshots(int n) {
		if (n >= 0) {
			numberOfSnapshots = n;
		}
		else {
			getLog().severe("Platform.setNumberOfSnapshots : invalid value " + n);
		}
	}
	
	public static void setRealTime(long clockUnitInMilliseconds, long intervalBetweenSnapshots) {
		Platform.clockUnitInMilliseconds = clockUnitInMilliseconds;
		Platform.intervalBetweenSnapshots  = intervalBetweenSnapshots;
	}

    static public void run() {
		//Clock clock = TimeFactory.eINSTANCE.createLogicalClock();
		//System.out.println("runClock = " + clock);
    	
		InstantValue zero = rootClock.createInstantValue(0);
		rootModel.setup(zero);
		
		startTimeInMilliseconds = System.currentTimeMillis();
		
		int snapshotsCount = 0;

	    boolean newsnap = true;
		while (newsnap) {
    		getLog().info("Platform.run: doing snapshot " + ++snapshotsCount);
    		controller.beginSnapshot(snapshotsCount);
	    	snapshot();
	    	newsnap = rootModel.newSnap();
	    	if ((numberOfSnapshots != 0) && (snapshotsCount >= numberOfSnapshots)) {
	    		getLog().info("Platform.run: all snapshots done");
	    		break;
	    	}
	    	controller.endSnapshot();
	    	
	    	// if running in real-time, the next snapshot is done after being requested
	    	if(clockUnitInMilliseconds > 0) {
	    		Thread timer = new Thread() {
	    			public void run() {
	    				try {
	    					long timeToWait;
	    					long currentTime = System.currentTimeMillis();
	    					if(dateOfLastSnapshot > 0) {
	    						if(currentTime - dateOfLastSnapshot <= intervalBetweenSnapshots)
	    							timeToWait = intervalBetweenSnapshots - (currentTime - dateOfLastSnapshot);
	    						else
	    							timeToWait = 0;
	    					} else timeToWait = 0;
	    					dateOfLastSnapshot = currentTime;
	    					
							sleep(timeToWait);
							// TODO this is a hack...
							// it only works on our example because SDF is sampled every clock unit
							snapshotRequested.release();
						} catch (InterruptedException e) {
							// if interrupted, just terminate gracefully
							// this thread is interrupted when the user requests
							// a snapshot through the UI
						}
	    			};
	    		};
	    		timer.start();
	    		
	    		snapshotRequested.acquireUninterruptibly();
	    		
	    		// terminate the timer thread if still running
	    		timer.interrupt();
	    	}
	    }
				
		rootModel.wrapup();
    }

     static private void snapshot() {
    	rootModel.startOfSnapshot();
    	
    	boolean snapOK = false;
    	while (!snapOK) {
    		if (!getTimeOfSnapshot()) break;
        	rootModel.reset(currentTime);
        	rootModel.computeSnapshot();
        	snapOK = rootModel.validate();
    	}

    	rootModel.endOfSnapshot();
    }
    
    static private boolean getTimeOfSnapshot() {
    	EList<InstantValue> timeConstraints = new BasicEList<InstantValue>();
    	rootModel.collectTimeConstraints(timeConstraints);

    	getLog().finer("Collected constraints: " + timeConstraints);
    	
    	InstantValue min = null;
    	InstantValue positiveMin = null;
    	for(InstantValue i : timeConstraints) {
    		if ((min == null) || (i.compareTo(min) < 0)) {
    			min = i;
    		}
    		
    		if ((i.getDenotedInstant().get(0).getDate() >= 0.) && ((positiveMin == null) || (i.compareTo(positiveMin) < 0))) {
    			//System.out.println(i + " < " + positiveMin);
    			positiveMin = i;
    		}
    	}
    	
    	if (min == null) {
    		getLog().info("Platform.getTimeOfSnapshot : no more time constraints");
    		return false;
    	}
    	
    	// If one block wants a snapshot ASAP, it posts an instant with value -1
    	if(min.getDenotedInstant().get(0).getDate() < 0) {
    		// We may handle it if running in real-time
    		if(clockUnitInMilliseconds > 0) {
    			double actualTime = ((double)(System.currentTimeMillis() - startTimeInMilliseconds)) / ((double)clockUnitInMilliseconds);
    			Platform.getLog().info("Platform requested to do a snapshot ASAP, will do it at " + actualTime);
    			//System.out.println("Platform requested to do a snapshot ASAP, will do it at " + actualTime);
    			min = min.getOnClock().createInstantValue(actualTime);
    			
    			// if a MoC had requested a snapshot "before the current real time"...
    			if(positiveMin != null && positiveMin.compareTo(min) < 0) {
    				Platform.getLog().info("Actually will do the snapshot at " + positiveMin + " because a MoC requested it");
    				System.err.println("Actually will do the snapshot at " + positiveMin + " and not " + min + " (which was supposed to be ASAP) because a MoC requested it");
    				min = positiveMin;
    			}
    			
    		} else throw new Error("Snapshot requested ASAP, supported only when running in real-time");
    	}
    	
    	// It is OK to have min.equals(currentTime), it means epsilon after currentTime
    	// TODO : modify the time package for that ??
    	if ((currentTime != null) && (min.compareTo(currentTime) < 0)) {
			System.err.println("Platform.getTimeOfSnapshot: new time (" + min + ") is not after the time of the previous snapshot (" + currentTime + ")");
    	}

    	getLog().fine("Platform.getTimeOfSnapshot : got time " + min.getDenotedInstant().get(0).getDate());
    	currentTime = min;
    	
    	// if running in real time, wait for the physical instant to come
    	if(clockUnitInMilliseconds > 0) {
    		long waitingTime = (long)(clockUnitInMilliseconds * currentTime.getDenotedInstant().get(0).getDate()) - (System.currentTimeMillis() - startTimeInMilliseconds);
    		if(waitingTime > 0)
	    		try {
	    			reportLag((clockUnitInMilliseconds - waitingTime) / ((double)clockUnitInMilliseconds));
					Thread.sleep(waitingTime);
				} catch (InterruptedException e) {}
    	}
    	
    	return true;
    }
    
    static public void requestSnapshot() {
    	snapshotRequested.release();
    }
    
    private static void reportLag(double lag) {
    	for(LagListener l : lagListeners)
    		l.updateLag(lag);
    }
    
    public static void addLagListener(LagListener l) {
    	lagListeners.add(l);
    }
    
    public static interface LagListener {
    	public void updateLag(double lag);
    }
}
