/**
 * Copyright (c) 2008-2011 Supélec.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * This file is part of the ModHel'X tool
 * 
 * Contributors:
 *     Computer Science Department, Supélec : initial implementation
 * Contacts:
 *     Cecile.Hardebolle@supelec.fr
 *     Christophe.Jacquet@supelec.fr
 *     Dominique.Marcadet@supelec.fr
 *     Frederic.Boulanger@supelec.fr
 *
 * $Id: FsmModelOfComputationImpl.java 331 2011-03-11 12:01:35Z marcadet $
 */
package fr.supelec.modhelx.mocs.fsm.impl;

import java.util.ArrayList;
import java.util.WeakHashMap;
import java.util.logging.Level;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;

import fr.supelec.modhelx.core.Block;
import fr.supelec.modhelx.core.CompositeBlock;
import fr.supelec.modhelx.core.CoreFactory;
import fr.supelec.modhelx.core.CorePackage;
import fr.supelec.modhelx.core.Model;
import fr.supelec.modhelx.core.Pin;
import fr.supelec.modhelx.core.Relation;
import fr.supelec.modhelx.core.Token;
import fr.supelec.modhelx.core.impl.ModelOfComputationImpl;
import fr.supelec.modhelx.core.time.InstantValue;
import fr.supelec.modhelx.engine.Platform;
import fr.supelec.modhelx.mocs.fsm.FsmModelOfComputation;
import fr.supelec.modhelx.mocs.fsm.FsmPackage;

/**
 * <!-- begin-user-doc -->
 * An implementation of the Finite State Machine '<em><b>Model Of Computation</b></em>'.
 * 
 * Limitations :
 * <ul>
 * <li>No automatic transition (i.e. without triggering event).</li>
 * <li>No internal events (i.e. produced by the machine and processed by the machine).</li>
 * <li>No entry/exit actions on states.</li>
 * </ul>
 * <!-- end-user-doc -->
 * <p>
 * </p>
 *
 * @generated
 */
public class FsmModelOfComputationImpl extends ModelOfComputationImpl implements FsmModelOfComputation {
	
	/**
	 * Current event to process in the snapshot
	 */
	private Token currentEvent = null;
	
	/**
	 * Current state of the automaton
	 */
	private Block currentState = null;
	
	/**
	 * Saved current state (for reset)
	 */
	private Block initialCurrentState = null;
	
	/**
	 * Transition which has been fired in the snapshot
	 */
	private Block executedTransition = null;
	
	/**
	 * List of the events to process.
	 * The events which are present on the input pins of the model at the beginning
	 * of a snapshot are put in that list.
	 * Each snapshot corresponds to the processing of one event (stored in the "currentEvent" variable
	 * in the start of snapshot).
	 */
	private ArrayList<Token> eventsToProcess = new ArrayList<Token>();
	
	/**
	 * List of the transitions which are fireable in the current snapshot.
	 * A transition is fireable in a snapshot if it is connected to the current state
	 * and is the target of an incoming event.
	 */
	private ArrayList<Block> fireableTransitions = new ArrayList<Block>();
	
	
	/**
	 * List of the pins of the transitions which have received an event
	 */
	private ArrayList<Pin> inputsOfFireableTransitionsWithEventReceived = new ArrayList<Pin>();
	
	/**
	 * List of the received events with their destination pin
	 */
	private WeakHashMap<Token, Pin> incommingPinOfEvents = new WeakHashMap<Token, Pin>();

	/**
	 * Variable for storing the current time received from the platform (no notion of time in FSM)
	 */
	private InstantValue currentTime = null;
	
	/**
	 * Roles of the different blocks for FSM
	 */
	private final static String
		FSM_ROLE_PARAM_NAME = "FSMRole",
		FSM_ROLE_STATE = "FSMState",
		FSM_ROLE_TRANSITION = "FSMTransition",
		FSM_ROLE_GUARD = "FSMGuard",
		FSM_ROLE_ACTION = "FSMAction";

	private final static String
		FSM_FEATURE_PARAM_NAME = "FSMFeature",
		FSM_FEATURE_INITIAL_STATE = "FSMInitialState";
	
	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	protected FsmModelOfComputationImpl() {
		super();
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated
	 */
	@Override
	protected EClass eStaticClass() {
		return FsmPackage.Literals.FSM_MODEL_OF_COMPUTATION;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void setup(Model model, InstantValue initialTime) {
		super.setup(model, initialTime);
		
		currentTime = initialTime;
		initialCurrentState = null;
		
		//Check that there is one and only one block with initial state role
		for (Block b : model.getStructure().getBlocks()) {
			if (hasInitialStateFeature(b)) {
				if (initialCurrentState != null) {
					throw new Error("Ill-formed model for FSM : more than 1 init block");					
				}
				initialCurrentState = b;
			}
		}
		
		if (initialCurrentState == null) {
			throw new Error("Ill-formed model for FSM : no init block");					
		}

		//TODO Perform other checks on the structure
		
		Platform.getLog().info("**** current state = " +initialCurrentState);

	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void startOfSnapshot(Model model) {
		super.startOfSnapshot(model);

		//Get all the events which are present on the input pins of the model
		// and store them in the list, along with their source pin
		for (Pin pinIn : model.getStructure().getPinsIn()) {
			if (pinIn.hasTokens()) {
				Token event = pinIn.getToken(true);
				eventsToProcess.add(event);
				incommingPinOfEvents.put(event, pinIn);
			}
		}
		
		Platform.getLog().info("**** eventsToProcess = " +eventsToProcess);

	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void endOfSnapshot(Model model) {
		super.endOfSnapshot(model);
		
		//Reinitialize the variables for the next snapshot
		executedTransition = null;

		//Put the events produced during the snapshot on the output pins of the model
		for (Block b : model.getStructure().getBlocks()) {
			for (Pin pinOut : b.getPinsOut()) {
				while (pinOut.hasTokens()) {
					Token t = pinOut.getToken(true);
					for (Relation r : pinOut.getOutgoingRelations()) {
						r.getTarget().addToken(t);
					}
				}
			}
		}
		
		// If the snapshot has been done (computeSnapshot), then
		// the first event of the list has been processed in the current snapshot
		// so it has to be removed from the list of the events to process
		if(eventsToProcess.size() > 0)
			eventsToProcess.remove(0);
		
		//Initialize the current state for the next snapshot
		initialCurrentState = currentState;
		
		if (Platform.getLog().isLoggable(Level.INFO)) {
			Platform.getLog().info("**** endOfSnapshot : new current state is " + initialCurrentState.getName());
		}
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void reset(Model model, InstantValue time) {
		super.reset(model, time);
		
		//FSM is passive, it receives its current time from the platform
		currentTime = time;

		//Clear all the variables which are used to know which blocks to update in the current snapshot
		fireableTransitions.clear();
		inputsOfFireableTransitionsWithEventReceived.clear();
		executedTransition = null;
		
		//Reinitialize the current state (saved in the end of snapshot)
		currentState = initialCurrentState;
		
		//The event to process is the first event of the list
		if(eventsToProcess.size()>0){
			currentEvent = eventsToProcess.get(0);
		}
		
		//We put tokens on the outputs of the current state in order to activate
		// the blocks (transitions, guards) which are connected to it
		for (Pin p : currentState.getPinsOut()) {
			p.addToken(CoreFactory.eINSTANCE.createToken());
//			p.addToken(currentEvent);
			//TODO CAUTION : pb if currentEvent is null 
			// Explanation : reset is called on the whole hierarachy of the model, so there may be no event to process at this time
			// This is particularly true on embedded FSM models (i.e. inside interface blocks)
		}
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void update(Model model) {
		// We don't update every block
		//super.update(model);
		Platform.getLog().entering(getClass().getName(), "update", model.getName());
		Platform.getController().enterMethod(this, "update", model);
		
		Platform.getLog().entering(getClass().getName(), "update", model.getName());

		//If in step 1 (= find the transition to fire)
		if (currentEvent != null) {

			//For each fireable transition, we check if its inside guard is positive (i.e. produces a token)
			int i = 0;
			for (Block b : fireableTransitions){
				if(!hasTransitionRole(b)) throw new Error("Ill-formed model for FSM : block "+b+" is not a transition");
				
				Platform.getLog().info("**** checked fireableTransition = " +b);

				//Get the input pin in the list of input pins which have received an event
				//TODO CAUTION : we assume that there is only one pin per transition and transitions 
				// and that pins are stored in the same order (done in initial schedule)
				Pin input = inputsOfFireableTransitionsWithEventReceived.get(i);
				++i;

				//If there is an inside guard in the transition (which is a Composite Block in this case)
				//TODO CAUTION : we assume that there is at least one outgoing relation (checked at the setup)
				if(input.getOutgoingRelations().size()>=1){
					Relation a = input.getOutgoingRelations().get(0);
					Pin dest = a.getTarget();

					//We put the token produced on pin of the inside guard to activate it 
					dest.addToken(currentEvent);

					//Then we update the inside guard
					Block guard = dest.getInputOf();
					if(!hasGuardRole(guard)) throw new Error("Ill-formed model for FSM : block "+guard+" is not a guard");
					guard.update(this);
					Platform.getLog().info("**** updated inside guard = " +guard);

					
					//And we check if it produces a token 
					//TODO CAUTION : we assume that a guard produces a token simultaneously on all of its output pins
					// therefore we only have to check one of them
					if (guard.getPinsOut().get(0).hasTokens()) {
						if (executedTransition != null) {
							throw new Error("FsmModelOfComputation.update : more than one successfull transition");					
						}
						executedTransition = b;
						Platform.getLog().info("**** positive guard = " +guard);
						Platform.getLog().info("**** executed transition = " + executedTransition);
					}
				}else{
					throw new Error("Ill-formed model for FSM : no inside guard in transition "+b);
					//TODO In another version of FSM : take into account the case of a transition without inside guard.
					//Two cases :
					// - automatic transition : taken without incoming event
					// - opaque transition : able to receive events but black box (produces an event if taken)
				}
			}

			Platform.getController().leaveMethod();

			return;
		}
		
		//Else, if in step 2 (= update the actions associated to the positive guard)
		if (executedTransition != null) {
			if(executedTransition.eClass().equals(CorePackage.eINSTANCE.getCompositeBlock())){
				CompositeBlock exectranscompo = (CompositeBlock)executedTransition;
				for (Block b : exectranscompo.getBlocks()) {
					if(hasActionRole(b)){
						b.update(this);
						Platform.getLog().info("**** updated action = " +b);
					}
				}
			}
		}
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public boolean snapDet(Model model) {
		super.snapDet(model);

		// If at the end of step 1 (= find the transition to fire) then executedTransition is not null, meaning that the snapshot is not finished 
		// If at the end of step 2 (= update the associated actions) then the snapshot is finished and executedTransition is null (done in the final schedule)
		return executedTransition == null;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public boolean newSnap(Model model) {
		super.newSnap(model);
		
		//If we have not processed all the events, a snapshot is needed
		if (! eventsToProcess.isEmpty()) return true;

		//If we have processed all the events in the list, we have to check if no new event
		// has been received on the input pins of the model
		for (Pin pinIn : model.getStructure().getPinsIn()) {
			if (pinIn.hasTokens()) {
				return true;
			}
		}
		
		//Else no new snapshot is needed
		return false;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void collectTimeConstraints(Model model, EList<InstantValue> timeConstraints) {
		//If a new snapshot is needed (events have to be processed), we post a constraint
		// at the current time
		if (newSnap(model)) {
			timeConstraints.add(currentTime);
		}
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void initialSchedule(Model model) {
		super.initialSchedule(model);
		
		Platform.getLog().info("FSM MoC test : initialschedule ");
		// Two steps : 
		// - the first to find the transition to fire (positive guard)
		// - the second to update associated action
		// TODO In another version of FSM : implement more steps (exit action, guard action, enter action)
		
		//If in step 1 (= find the transition to fire)
		if (currentEvent != null) {
			//We list all the fireable transitions, i.e. all the blocks which have transition role,
			// are the target of an incoming event AND are connected to the current state

			//For all the relations which leave the pin which has received the current event
			for (Relation r : incommingPinOfEvents.get(currentEvent).getOutgoingRelations()) {
				//Get the connected block and check its role
				Block b = r.getTarget().getInputOf();
				//If it is a transition
				if(hasTransitionRole(b)){
					for (Pin p : b.getPinsIn()) {
						for (Relation r2 : p.getIncomingRelations()) {
							//And if it is connected to the current state
							if (r2.getSource().getOutputOf() == currentState) {
								//We add it to the list of the fireable transitions
								fireableTransitions.add(b);
								//And we store its input pin in the list of inputs of fireable transitions
								// which have received an event
								inputsOfFireableTransitionsWithEventReceived.add(r.getTarget());
								break;
							}
						}
					}
				}
			}
			Platform.getLog().info("**** fireableTransitions = " +fireableTransitions);

			return;
		}
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void prePropagate(Model model) {
		super.prePropagate(model);
		
		//If in step 1 (= find the fireable transition)
		if (currentEvent != null) {
			//For all the output pins of the current state
			for (Pin p : currentState.getPinsOut()) {
				//If events have been produced
				if (p.hasTokens()) {
					Token t = p.getToken(true);
					//We move them on the connected pins (i.e. pins of transitions)
					for (Relation r : p.getOutgoingRelations()) {
						r.getTarget().addToken(t);
					}
				}
			}
			
			//For all the fireable transitions
			for (Block b : fireableTransitions) {
				for (Pin p : b.getPinsIn()) {
					//If events have been put on the input of a fireable transition
					// and this fireable transition has an inside guard
					if (p.hasTokens() && p.getOutgoingRelations().size()>=1) {
						//We move the events inside the transition on the input of the connected guards
						Token t = p.getToken(true);
						for (Relation r : p.getOutgoingRelations()) {
							//TODO CAUTION : this check may be annoying if we allow an action to be the target of an event
							if(!hasGuardRole(r.getTarget().getInputOf())) throw new Error("Ill-formed model for FSM : block "+r.getTarget().getInputOf()+" is not a guard");
							r.getTarget().addToken(t);
						}
					}
				}
			}
			return;
		}
		
		//Else, if in step 2 (= update the associated actions) 
		if (executedTransition != null) {
			//If the transition has inside guards and actions, 
			if(executedTransition.eClass().equals(CorePackage.eINSTANCE.getCompositeBlock())){
				CompositeBlock exectranscompo = (CompositeBlock)executedTransition;
				for (Block b : exectranscompo.getBlocks()) {
					for (Pin p : b.getPinsOut()) {
						while (p.hasTokens()) {
							//We propagate the tokens produced by the guard to the connected actions
							Token t = p.getToken(true);
							for (Relation r : p.getOutgoingRelations()) {
								r.getTarget().addToken(t);
							}
						}
					}
				}
			}
			//Else nothing to propagate
		}
	}


	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void postPropagate(Model model) {
		super.postPropagate(model);
		
		//If in step 1 (= find the fireable transition)
		if (currentEvent != null) {
			//nothing to do
			return;
		}
		
		//Else, if in step 2 (= update the associated actions) 
		if (executedTransition != null) {
			//If the transition has inside guards and actions, 
			if(executedTransition.eClass().equals(CorePackage.eINSTANCE.getCompositeBlock())){
				CompositeBlock exectranscompo = (CompositeBlock)executedTransition;
				for (Block b : exectranscompo.getBlocks()) {
					for (Pin p : b.getPinsOut()) {
						while (p.hasTokens()) {
							//We propagate the tokens produced by the action to the connected output pins of the composite
							Token t = p.getToken(true);
							for (Relation r : p.getOutgoingRelations()) {
								r.getTarget().addToken(t);
							}
						}
					}
				}
			}
			//If the transition has tokens on its output pins (eventually produced by its inside actions)
			for (Pin p : executedTransition.getPinsOut()) {
				while (p.hasTokens()) {
					Token t = p.getToken(true);
					for (Relation r : p.getOutgoingRelations()) {
						//We move the tokens to the connected pins
						r.getTarget().addToken(t);

						//If the connected pin is an input, that means that this block is the new current state
						if (r.getTarget().getOutputOf() == null) {
							currentState = r.getTarget().getInputOf();
							Platform.getLog().info("**** new current state = " +currentState);
						}
					}
				}
			}
		}
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void finalSchedule(Model model) {
		super.finalSchedule(model);
		// Two steps : 
		// - the first to find the positive guard
		// - the second to update associated action
		// TODO In another version of FSM : implement more steps (exit action, guard action, enter action)
		
		//If in step 1 (= find the fireable transition)
		if (currentEvent != null) {
			currentEvent = null; //We move in step 2
			return;
		}
		
		//Else, if in step 2 (= update the associated actions) 
		if (executedTransition != null) {
			executedTransition = null; //We move in step 1 for the next snapshot
		}
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void setStateRole(Block block) {
		block.setValueOfParameter(FSM_ROLE_PARAM_NAME, FSM_ROLE_STATE);
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public boolean hasStateRole(Block block) {
		if(block.getParameters().containsKey(FSM_ROLE_PARAM_NAME)){
			return block.getValueOfParameter(FSM_ROLE_PARAM_NAME).equals(FSM_ROLE_STATE);
		}
		return false;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void setGuardRole(Block block) {
		block.setValueOfParameter(FSM_ROLE_PARAM_NAME, FSM_ROLE_GUARD);
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public boolean hasGuardRole(Block block) {
		if(block.getParameters().containsKey(FSM_ROLE_PARAM_NAME)){
			return block.getValueOfParameter(FSM_ROLE_PARAM_NAME).equals(FSM_ROLE_GUARD);
		}
		return false;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void setActionRole(Block block) {
		block.setValueOfParameter(FSM_ROLE_PARAM_NAME, FSM_ROLE_ACTION);
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public boolean hasActionRole(Block block) {
		if(block.getParameters().containsKey(FSM_ROLE_PARAM_NAME)){
			return block.getValueOfParameter(FSM_ROLE_PARAM_NAME).equals(FSM_ROLE_ACTION);
		}
		return false;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void setInitialStateFeature(Block block) {
		if(!block.getParameters().containsKey(FSM_ROLE_PARAM_NAME)){
			block.setValueOfParameter(FSM_ROLE_PARAM_NAME, FSM_ROLE_STATE);
		}else if(block.getParameters().containsKey(FSM_ROLE_PARAM_NAME) && !hasStateRole(block)){
			throw new Error("Ill-formed model for FSM : cannot set the role INIT_STATE to a block which does not have STATE role");
		}		
		block.setValueOfParameter(FSM_FEATURE_PARAM_NAME, FSM_FEATURE_INITIAL_STATE);
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public boolean hasInitialStateFeature(Block block) {
		return (block.getParameters().containsKey(FSM_FEATURE_PARAM_NAME) && block.getValueOfParameter(FSM_FEATURE_PARAM_NAME).equals(FSM_FEATURE_INITIAL_STATE));
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void setTransitionRole(Block block) {
		block.setValueOfParameter(FSM_ROLE_PARAM_NAME, FSM_ROLE_TRANSITION);
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public boolean hasTransitionRole(Block block) {
		if(block.getParameters().containsKey(FSM_ROLE_PARAM_NAME)){
			return block.getValueOfParameter(FSM_ROLE_PARAM_NAME).equals(FSM_ROLE_TRANSITION);
		}
		return false;
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public Token createEvent() {
		return CoreFactory.eINSTANCE.createToken();
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void startOfUpdate(Model model) {
		super.startOfUpdate(model);
		
		Platform.getLog().info("FSM MoC test : start of update ");
		//Get all the events which are present on the input pins of the model
		// and store them in the list, along with their source pin
		for (Pin pinIn : model.getStructure().getPinsIn()) {
			if (pinIn.hasTokens()) {
				Token event = pinIn.getToken(true);
				eventsToProcess.add(event);
				incommingPinOfEvents.put(event, pinIn);
			}
		}
		
		//The event to process is the first event of the list
		if(eventsToProcess.size()>0) currentEvent = eventsToProcess.get(0);
		
		Platform.getLog().info("**** eventsToProcess = " +eventsToProcess);
	}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public void endOfUpdate(Model model) {
		super.endOfUpdate(model);
		
		Platform.getLog().info("FSM MoC test : end of update (empty)");

		//Reinitialize the variables for the next update
		executedTransition = null;

		//Put the events produced during the snapshot on the output pins of the model
		for (Block b : model.getStructure().getBlocks()) {
			for (Pin pinOut : b.getPinsOut()) {
				while (pinOut.hasTokens()) {
					Token t = pinOut.getToken(true);
					for (Relation r : pinOut.getOutgoingRelations()) {
						r.getTarget().addToken(t);
					}
				}
			}
		}
		
		// If the snapshot has been done (computeSnapshot), then
		// the first event of the list has been processed in the current snapshot
		// so it has to be removed from the list of the events to process
		if(eventsToProcess.size() > 0)
			eventsToProcess.remove(0);
		
		//Initialize the current state for the next snapshot
		initialCurrentState = currentState;
		
		if (Platform.getLog().isLoggable(Level.INFO)) {
			Platform.getLog().info("**** endOfSnapshot : new current state is " + initialCurrentState.getName());
		}
}

	/**
	 * <!-- begin-user-doc -->
	 * <!-- end-user-doc -->
	 * @generated NOT
	 */
	public boolean further(Model model) {
		super.further(model);
		Platform.getLog().info("FSM MoC test : further");
		
		// If at the end of step 1 (= find the transition to fire) then executedTransition is not null, meaning that the snapshot is not finished 
		return executedTransition != null;
//		return !this.snapDet(model);
	}

} //FsmModelOfComputationImpl
