package fr.supelec.modhelx.qdgui;

import java.awt.Color;
import java.awt.Graphics;
import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.emf.ecore.EObject;

import fr.supelec.graphbench.CompositeVertex;
import fr.supelec.graphbench.Edge;
import fr.supelec.graphbench.Socket;
import fr.supelec.graphbench.Vertex;
import fr.supelec.graphbench.ui.PropertyDescriptorProvider;
import fr.supelec.modhelx.core.AtomicBlock;
import fr.supelec.modhelx.core.Block;
import fr.supelec.modhelx.core.CompositeBlock;
import fr.supelec.modhelx.core.CoreFactory;
import fr.supelec.modhelx.core.InterfaceBlock;
import fr.supelec.modhelx.core.ModelOfComputation;
import fr.supelec.modhelx.core.Pin;
import fr.supelec.modhelx.core.Relation;
import fr.supelec.modhelx.qdgui.rendering.RendererParameters;
import fr.supelec.modhelx.qdgui.rendering.VertexRenderer;
import fr.supelec.modhelx.qdgui.rendering.VertexRenderers;
import fr.supelec.modhelx.qdgui.rendering.VertexRenderer.RenderInternals;

public class CompositeVertexAdapter extends CompositeVertex implements VertexAdapter, PropertyDescriptorProvider {
	public static final int RESIZE_PAD_SIZE = 10;
	private final Block block;
	private final List<Vertex> vertices = new ArrayList<Vertex>();
	private final List<Edge> edges = new ArrayList<Edge>();
	private List<SocketAdapter> sockets = new ArrayList<SocketAdapter>();
	private VertexRenderer customRenderer;
	private Color backgroundColor = Color.WHITE;
	private final static Color InterfaceBlockColor = new Color(220, 220, 220);
	private final RendererParameters rendererParams;
	private final ModelOfComputation moc;
	
	/**
	 * 
	 * @param parent
	 * @param block a composite or interface block
	 */
	public CompositeVertexAdapter(CompositeVertex parent, Block block, ModelOfComputation moc) {
		super(parent, 10, 10);
		//super(parent, parent == null ? 1000 : 400, parent == null ? 600 : 300);
		
		this.block = block;
		this.moc = moc;
		
		customRenderer = VertexRenderers.forBlock(block);
		// at this point, customRenderer may not be null
		if(customRenderer == null) throw new Error("Null renderer for block " + block);
		
		Set<Block> placedBlocks = new HashSet<Block>(); 
		
		// first place source blocks on the left
		int x = 30;
		int y = 30;
		int xMax = 0;
		
		rendererParams = customRenderer.calculateRendererParameters(this);
		
		Common.setSockets(block, this, sockets, customRenderer.renderPinNames());
				
		if(block instanceof CompositeBlock) {
			CompositeBlock compositeBlock = (CompositeBlock) block;
			int yMax = 0;
			
			// place internal blocks and then resize the composite blocks
			// accordingly if its internals are to be rendered
			if(customRenderer.renderInternals().atLeast(RenderInternals.RENDER_BLOCKS)) {
				// under them, source blocks
				for(Block b : compositeBlock.getBlocks()) {
					if(b.getPinsIn().size() > 0) continue;  // skip any non-source block
					
					///System.out.println("Source: " + b);
					Vertex va = createVertexAdapter(b);
					vertices.add(va);
					va.setPosition(x, y);
					y += 40 + va.getHeight();
					if(y > yMax) yMax = y;
					placedBlocks.add(b);
					if(x + va.getWidth() > xMax) xMax = x + va.getWidth();
				}
				
				// then place the other blocks in the middle
				if(y>60) y = y/2 - 30;
				x = xMax + 60; 
				for(Block b : compositeBlock.getBlocks()) {
					if(placedBlocks.contains(b)) continue;  // skips already-placed blocks
					///System.out.println(b);
					Vertex va = createVertexAdapter(b);
					vertices.add(va);
					va.setPosition(x, y);
					x += 60 + va.getWidth();
					// botom of the block + 40px margin
					int yBottom = y + va.getHeight() + 40;
					if(yBottom > yMax) yMax = yBottom;
				}
			}

			// do it even if the internals are not to be rendered: it will
			// cause the adapter to know the renderer dimensions
			setWidth(x);
			setHeight(yMax);

			
			// determine the hierarchical level: the color will be darker
			// if components are imbricated, so as to help distinguish between
			// them
			int level = 0;
			EObject cur = block;
			while(cur.eContainer() instanceof CompositeBlock) {
				level++;
				cur = cur.eContainer();
			}
			
			if(level > 0) backgroundColor = new Color(1f - .1f * level, 1f, 1f - .1f * level);
			
		} else if(block instanceof InterfaceBlock) {
			InterfaceBlock interfaceBlock = (InterfaceBlock) block;
			
			Vertex va = createVertexAdapter(interfaceBlock.getInternalModel().getStructure());
			vertices.add(va);
			va.setPosition(x, y);
			setWidth(va.getWidth() + 2*x);
			setHeight(va.getHeight() + 2*y);
			backgroundColor = InterfaceBlockColor;
		} else throw new Error("CompositeVertexAdapter only exists for CompositeBlock and InterfaceBlock.");
		
		updateEdges();
	}


	@Override
	public void addEdge(Socket from, Socket to) {
		if(block instanceof CompositeBlock) {
			CompositeBlock compositeBlock = (CompositeBlock) block;
			SocketAdapter froma = (SocketAdapter)from;
			SocketAdapter toa = (SocketAdapter)to;
			Relation rel = CoreFactory.eINSTANCE.createRelation();
			rel.setSource(froma.getPin());
			rel.setTarget(toa.getPin());
			compositeBlock.getRelations().add(rel);
			edges.add(new EdgeAdapter(this, rel));  // update the shadow model
		} // TODO implement it for InterfaceBlock?
	}

	@Override
	public void addVertex(Vertex vertex) {
		if(block instanceof CompositeBlock) {
			CompositeBlock compositeBlock = (CompositeBlock) block;
			AtomicVertexAdapter va = (AtomicVertexAdapter)vertex;
			vertices.add(va);
			compositeBlock.getBlocks().add(va.getBlock());
		} // TODO implement it for InterfaceBlock?
	}

	@Override
	public Iterable<Edge> edges() {
		return edges;
	}

	@Override
	public boolean removeEdge(Edge edge) {
		if(block instanceof CompositeBlock) {
			CompositeBlock compositeBlock = (CompositeBlock) block;
			edges.remove(edge);   // update the shadow model
			Relation rel = ((EdgeAdapter)edge).getRelation();
			
			// update the ModHel'X model
			compositeBlock.getRelations().remove(rel);
			rel.getSource().getOutgoingRelations().remove(rel);
			rel.getTarget().getIncomingRelations().remove(rel);
			return true;
		} else return false; // TODO implement it for InterfaceBlock?
	}

	@Override
	public boolean removeVertex(Vertex vertex) {
		if(block instanceof CompositeBlock) {
			CompositeBlock compositeBlock = (CompositeBlock) block;
			
			// delete connected edges
			removeAfferentEdges(vertex);
			
			vertices.remove(vertex);
			return compositeBlock.getRelations().remove(((AtomicVertexAdapter)vertex).getBlock());
		} else return false; // TODO implement it for InterfaceBlock?
	}

	@Override
	public Iterable<Vertex> vertices() {
		return vertices;
	}
	
	public SocketAdapter getSocketForPin(Pin pin) {
		//Block block = pin.getInputOf();
		//if(block == null) block = pin.getOutputOf();
		
		// first, does this pin belong to the composite block?
		for(SocketAdapter s : sockets) {
			if(s.getPin() == pin) return s;
		}

		
		// then, try to find this pin as an I/O of one of the contained blocks
		for(Vertex v : vertices()) {
			Socket s = null;
			
			if(v instanceof CompositeVertexAdapter) {
				s = ((CompositeVertexAdapter)v).getSocketForPin(pin);
			} else if(v instanceof AtomicVertexAdapter) {
				s = ((AtomicVertexAdapter)v).getSocketForPin(pin);
			}
			if(s != null) return (SocketAdapter)s;
		}
		return null;
	}
	
	private void updateEdges() {
		// nothing to do if the internals are not to be rendered
		if(! customRenderer.renderInternals().atLeast(RenderInternals.RENDER_BLOCKS_AND_INTER_BLOCK_RELATIONS)) return;
		
		List<Relation> relations;
		if(block instanceof CompositeBlock) {
			relations = ((CompositeBlock) block).getRelations();
		} else if(block instanceof InterfaceBlock) {
			relations = ((InterfaceBlock) block).getRelations();
		} else throw new Error("In CmpositeVertexAdapter, the block must be either InterfaceBlock or CompositeBlock.");
		
		for(Relation r : relations) {
			if(!isInterfaceRelation(r) || customRenderer.renderInternals().atLeast(RenderInternals.RENDER_ALL)) {
				EdgeAdapter ea = new EdgeAdapter(this, r);
				edges.add(ea);
			}
		}
	}
	
	private boolean isInterfaceRelation(Relation r) {
		return 
			block.getPinsIn().contains(r.getSource()) ||
			block.getPinsOut().contains(r.getTarget());
	}

	@Override
	public Iterable<? extends Socket> sockets() {
		return sockets;
	}
	
	public List<SocketAdapter> getSocketList() {
		return sockets;
	}

	@Override
	public void render(Graphics g) {
		int x1 = getX();
		int y1 = getY();
		
		customRenderer.render(rendererParams, g, x1, y1, isSelected());
		
		for(Socket s : sockets) {
			s.render(g);
		}
		
		if(customRenderer.renderInternals().atLeast(RenderInternals.RENDER_BLOCKS))
			renderInternals(g);

	}

	@Override
	public void writeElementAttributes(PrintStream s) {

	}


	@Override
	public CompositeVertex vertexToResize(int x, int y) {
		final int x2 = getX()+getWidth(), y2 = getY()+getHeight();
		if(x >= x2-RESIZE_PAD_SIZE && y >= y2-RESIZE_PAD_SIZE &&x <= x2 && y <= y2)
			return this;
		
		for(Vertex v : vertices()) {
			CompositeVertex vtr = v.vertexToResize(x - getX(), y - getY());
			if(vtr != null) return vtr;
		}
		
		return null;
	}
	
	@Override
	public void setWidth(int width) {
		// do not set size if the internals are not to be rendered
		if(customRenderer.renderInternals().atLeast(RenderInternals.RENDER_BLOCKS)) 
			rendererParams.width = width;
		
		customRenderer.recalculatePortPositions(rendererParams);
		
		super.setWidth(rendererParams.width);
	}
	
	@Override
	public void setHeight(int height) {
		// do not set size if the internals are not to be rendered
		if(customRenderer.renderInternals().atLeast(RenderInternals.RENDER_BLOCKS)) 
			rendererParams.height = height;
		
		customRenderer.recalculatePortPositions(rendererParams);
		
		super.setHeight(rendererParams.height);
	}
	
	// TODO: the same as above goes for setHeight, if there are sockets
	// attached to the lower edge of the component
	
	private Vertex createVertexAdapter(Block b) {
		if(b instanceof AtomicBlock) {
			return new AtomicVertexAdapter(this, b);
		} else if(b instanceof CompositeBlock) {
			return new CompositeVertexAdapter(this, b, moc);
		} else if(b instanceof InterfaceBlock) {
			return new CompositeVertexAdapter(this, b, ((InterfaceBlock)b).getInternalModel().getMoc());
		}  else {
			throw new Error("Cannot create an adapter for a block which is neither atomic nor composite nor an interface block. Block " + b);
		}
	}
	
	public void fitSize() {
		// do not fit size if the internals are not to be rendered
		if(!customRenderer.renderInternals().atLeast(RenderInternals.RENDER_BLOCKS)) return;
		
		int xMax = 0, yMax = 0;
		for(Vertex va : vertices) {
			xMax = Math.max(va.getX() + va.getWidth(), xMax);
			yMax = Math.max(va.getY() + va.getHeight(), yMax);
		}
		
		setWidth(xMax + SocketAdapter.SPACING);
		setHeight(yMax + SocketAdapter.SPACING);
	}
	

	public PropertyDescriptor[] getPropertyDescriptors() {
		try {
			BeanInfo bi = Introspector.getBeanInfo(block.getClass());
			return bi.getPropertyDescriptors();
		} catch(IntrospectionException e) {
			// TODO
			return new PropertyDescriptor[] {};
		}
	}
	
	public Object getPropertyHolder() {
		return block;
	}

	public String toString() {
		return "Composite " + block.getName();
	}
	
	@Override
	public String getToolTip() {
    	return Common.getToolTip(block, null);
	}
	
	public Block getBlock() {
		return block;
	}
	
    public StringBuffer serializePlacement() {
    	StringBuffer resEdges = new StringBuffer();
    	for(Edge e : edges) {
    		StringBuffer ee = ((EdgeAdapter)e).serializePlacement();
    		if(ee != null)
    			resEdges.append(ee).append("\n");
    	}
    	
    	StringBuffer res = Common.serializeVertexPlacement(this, resEdges);
    	res.append("\n<!-- Innards of " + block.getName() + " -->\n");
    	for(Vertex v : vertices) {
    		res.append(((VertexAdapter)v).serializePlacement()).append("\n");
    	}
    	
    	res.append("<!-- Done with the innards of " + block.getName() + " -->");
    	return res;
    }

    
    public Color getBackgroundColor() {
    	return backgroundColor;
    }
    
    public ModelOfComputation getMoc() {
    	return moc;
    }
}
