package fr.supelec.modhelx.tests;

import java.util.logging.Level;
import java.util.logging.SimpleFormatter;

import fr.supelec.modhelx.core.Block;
import fr.supelec.modhelx.core.CompositeBlock;
import fr.supelec.modhelx.core.CoreFactory;
import fr.supelec.modhelx.core.Model;
import fr.supelec.modhelx.core.Pin;
import fr.supelec.modhelx.core.time.Clock;
import fr.supelec.modhelx.core.time.TimeFactory;
import fr.supelec.modhelx.engine.Platform;
import fr.supelec.modhelx.mocs.sdf.SdfFactory;
import fr.supelec.modhelx.mocs.sdf.SdfModelOfComputation;
import fr.supelec.modhelx.mocs.sdflib.SdfPrinter;
import fr.supelec.modhelx.mocs.sdflib.SdflibFactory;
import fr.supelec.modhelx.qdgui.ModelWindow;

/**
 * This test of the SDF delay computes the Fibonacci series
 * using an adder in a loop with two delays. The delays are
 * used to memorize the last two values of the series and
 * feed the adder with them:
 * 
 *       +-------------+
 *       |             |
 *   +---|  Delay {0}  |<---+
 *   |   |             |    |
 *   |   +-------------+    |
 *   |                      |
 *   |   +-------------+    |        +---------+
 *   +-->|             |----+        |         |
 *       |    Adder    |------------>| Printer |
 *   +-->|             |----+        |         |
 *   |   +-------------+    |        +---------+
 *   |                      |
 *   |   +-------------+    |
 *   |   |             |    |
 *   +---| Delay {0,1} |<---+
 *       |             |
 *       +-------------+
 *       
 * @author Frédéric Boulanger
 *
 */
public class TestSDF_Fibonacci {

	private static CoreFactory coreFactory = CoreFactory.eINSTANCE;
	private static TimeFactory timeFactory = TimeFactory.eINSTANCE;
	private static SdfFactory sdfFactory = SdfFactory.eINSTANCE;
	private static SdflibFactory sdflibFactory = SdflibFactory.eINSTANCE;
	
	private static Model model = coreFactory.createModel();
	
	private static Clock clock;

	public static void main(String[] args) {
		buildModel();

		Platform.setLogLevel(Level.FINEST);
		Platform.setLogFormatter(new SimpleFormatter());
		Platform.setNumberOfSnapshots(0);

		Platform.setModel(model, clock);

		ModelWindow w = new ModelWindow(model, clock, "tests/TestSDF_Fibonacci.xy");
		w.setVisible(true);
		w.run();

		Platform.run();
	}

	private static void buildModel() {
		model.setName("TestSDF_Fibonacci");

		// MoC
		SdfModelOfComputation moc = sdfFactory.createSdfModelOfComputation();
		moc.setName("TestSDF_Fibonacci_MoC");
		clock = TimeFactory.eINSTANCE.createLogicalClock();
		moc.setClock(clock);
		model.setMoc(moc);

		// Structure
		CompositeBlock modelStructure = coreFactory.createCompositeBlock();
		modelStructure.setName("TestSDF_Fibonacci_Structure");
		model.setStructure(modelStructure);

		// Adder
		Block add = modelStructure.addBlock(sdflibFactory.createSdfAdder(), "add");
		Pin fromDelay1 = add.newInputPin("fromDelay1");
		Pin fromDelay2= add.newInputPin("fromDelay2");
		Pin toPrinter= add.newOutputPin("toPrinter");
		Pin toDelay1= add.newOutputPin("toDelay1");
		Pin toDelay2= add.newOutputPin("toDelay2");

		// Delays
		Block delay1 = modelStructure.addBlock(sdflibFactory.createSdfDelay(), "delay1");
		delay1.setValueOfParameter("initialData", 0);
		Pin inDelay1 = delay1.newInputPin("inDelay1");
		Pin outDelay1 = delay1.newOutputPin("outDelay1");
		delay1.setValueOfParameter("initialData", 0);

		Block delay2 = modelStructure.addBlock(sdflibFactory.createSdfDelay(), "delay2");
		delay2.setValueOfParameter("initialData", 0);
		Pin inDelay2 = delay2.newInputPin("inDelay2");
		Pin outDelay2 = delay2.newOutputPin("outDelay2");
		Integer init[] = {0, 1};
		delay2.setValueOfParameter("initialData", init);
		
		//Printer
		Block printer = modelStructure.addBlock(sdflibFactory.createSdfPrinter(), "sdfPrinter");
		((SdfPrinter) printer).setPrintStream(System.out);
		Pin inPrinter = printer.newInputPin("inPrinter");

		//Relations
		modelStructure.connect(toDelay1, inDelay1);
		modelStructure.connect(toDelay2, inDelay2);
		modelStructure.connect(outDelay1, fromDelay1);
		modelStructure.connect(outDelay2, fromDelay2);
		modelStructure.connect(toPrinter, inPrinter);

	}
}
