/**
 * Copyright (c) 2008-2011 Supélec.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * This file is part of the ModHel'X tool
 * 
 * Contributors:
 *     Computer Science Department, Supélec : initial implementation
 * Contacts:
 *     Cecile.Hardebolle@supelec.fr
 *     Christophe.Jacquet@supelec.fr
 *     Dominique.Marcadet@supelec.fr
 *     Frederic.Boulanger@supelec.fr
 *
 * $Id: TestSimpleMoC.java 331 2011-03-11 12:01:35Z marcadet $
 */

package fr.supelec.modhelx.tests;

import java.io.File;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.SimpleFormatter;

import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.xmi.impl.XMIResourceFactoryImpl;

import fr.supelec.modhelx.core.CompositeBlock;
import fr.supelec.modhelx.core.CoreFactory;
import fr.supelec.modhelx.core.Model;
import fr.supelec.modhelx.core.Pin;
import fr.supelec.modhelx.core.Relation;
import fr.supelec.modhelx.core.time.Clock;
import fr.supelec.modhelx.core.time.TimeFactory;
import fr.supelec.modhelx.engine.Platform;
import fr.supelec.modhelx.mocs.simple.SimpleConsumer;
import fr.supelec.modhelx.mocs.simple.SimpleFactory;
import fr.supelec.modhelx.mocs.simple.SimpleModelOfComputation;
import fr.supelec.modhelx.mocs.simple.SimpleProducer;

public class TestSimpleMoC {
	
	private static Model model = CoreFactory.eINSTANCE.createModel();
	
	public static void main(String[] args) {
		
		buildModel();
		saveModel();
		model = null;
		loadModel();

		Platform.setLogFile("log/simple_log.txt");
		Platform.setLogLevel(Level.FINEST);
		Platform.setLogFormatter(new SimpleFormatter());
		
		Platform.setModel(model, TimeFactory.eINSTANCE.createLogicalClock());
		
		Platform.run();
	}
	
	/*
	 *   +---------------------------------------------+
	 *   | M1                                          |
	 *   |  +----------------------------------------+ |
	 *   |  | CB2                   +--------------+ | |
	 *   |  |                       | CB4          | | |
	 *   |  |  +----------+         |    +-------+ | | |
	 *   |  |  | AB3      |         |    | AB5   | | | |
	 *   |  |  |          |         |    |       | | | |
	 *   |  |  |          >--------->---->       | | | |
	 *   |  |  |       OP3|         |IP4 |IP5    | | | |
	 *   |  |  +----------+         |    |       | | | |
	 *   |  |                       |    +-------+ | | |
	 *   |  |                       +--------------+ | |
	 *   |  +----------------------------------------+ |
	 *   +---------------------------------------------+
	 */
	private static void buildModel() {
		model.setName("M1");

		SimpleModelOfComputation moc = SimpleFactory.eINSTANCE.createSimpleModelOfComputation();
		moc.setName("MOC1");
		model.setMoc(moc);
		
		CompositeBlock cb2 = CoreFactory.eINSTANCE.createCompositeBlock();
		cb2.setName("CB2");
		model.setStructure(cb2);

		SimpleProducer ab3 = SimpleFactory.eINSTANCE.createSimpleProducer();
		ab3.setName("AB3");
		cb2.getBlocks().add(ab3);
		
		// The out Pin must not be created in SimpleProducer constructor, otherwise
		// it is duplicated when the model is loaded from XMI (one
		// when the SimpleProducer is created, and one in the XMI file !).
		Pin op3 = CoreFactory.eINSTANCE.createPin();
		op3.setName("OP3");
		ab3.getPinsOut().add(op3); 

		CompositeBlock cb4 = CoreFactory.eINSTANCE.createCompositeBlock();
		cb4.setName("CB4");
		cb2.getBlocks().add(cb4);
		
		Pin ip4 = CoreFactory.eINSTANCE.createPin();
		ip4.setName("IP4");
		cb4.getPinsIn().add(ip4);

		SimpleConsumer ab5 = SimpleFactory.eINSTANCE.createSimpleConsumer();
		ab5.setName("AB5");
		cb4.getBlocks().add(ab5);
		
		Pin ip5 = CoreFactory.eINSTANCE.createPin();
		ip5.setName("IP5");
		ab5.getPinsIn().add(ip5); 

		Relation op3_ip4 = CoreFactory.eINSTANCE.createRelation();
		op3_ip4.setSource(op3);
		op3_ip4.setTarget(ip4);
		cb2.getRelations().add(op3_ip4);
		
		//PinAlias ip4_ip5 = CoreFactory.eINSTANCE.createPinAlias();
		Relation ip4_ip5 = CoreFactory.eINSTANCE.createRelation();
		//ip4_ip5.setFrom(ip4);
		//ip4_ip5.setTo(ip5);
		ip4_ip5.setSource(ip4);
		ip4_ip5.setTarget(ip5);
		//cb4.getAliases().add(ip4_ip5);
		cb4.getRelations().add(ip4_ip5);
	}

	private static void saveModel() {
		// create resource set
		ResourceSet resourceSet = new ResourceSetImpl();
		// Register XML resource factory (before creating the resource)
		resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put(
				"modhelx", new XMIResourceFactoryImpl());
		Resource resource = resourceSet.createResource(URI.createFileURI(
				new File("simple.modhelx").getAbsolutePath()));
		// add the root object to the resource
		resource.getContents().add(model);
		// serialize resource
		try {
			resource.save(null);
		}
		catch(IOException ex) {
			System.err.println("Exception : " + ex);
		}
	}
	
	private static boolean loadModel() {
		ResourceSet resourceSet = new ResourceSetImpl();
		resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put(
				"modhelx", new XMIResourceFactoryImpl());
		Resource resource = resourceSet.createResource(URI.createFileURI(
				new File("simple.modhelx").getAbsolutePath()));
		try {
			resource.load(null);
		}
		catch(IOException ex) {
			System.err.println("Exception : " + ex);
			return false;
		}
		
		if (resource.getContents().size() != 1) {
			System.err.println("Error : there must be 1 element");
			return false;
		}
		
		EObject o = resource.getContents().get(0);
		try {
			model = (Model) o;
		}
		catch(ClassCastException ex) {
			System.err.println("Error : expected element not found : " + ex);
			return false;
		}
		return true;
	}
}
