package fr.supelec.modhelx.util;

import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import fr.supelec.modhelx.core.Block;
import fr.supelec.modhelx.core.CompositeBlock;
import fr.supelec.modhelx.core.Pin;
import fr.supelec.modhelx.core.Relation;
import fr.supelec.modhelx.engine.Platform;

public class DEBlockDependencyComparator implements Comparator<Block> {

	private final static String	DE_NO_INSTANTANEOUS_DEPENDENCY = "DENoInstantaneousDependency";

	/**
	 * Compare two blocks according to their dependencies.
	 * The relation of dependency between blocks is the transitive 
	 * closure of the relations between pins of the blocks.
	 * Some pins may be marked as "not instantaneously dependent". 
	 * Relations which have such a pin as source or target are
	 * ignored when building the graph of dependencies between blocks.
	 * A "not instantaneously dependent" output is an output with
	 * a state determined value (it does not depend on the current 
	 * inputs of the block). A "not instantaneously dependent" input 
	 * is an input which determines only the next state of the block, 
	 * not its current outputs.
	 * 
	 * @return 1 if b2 depends on b1 (b1 is before b2)
	 *        -1 if b1 depends on b2 (b1 is after b2)
	 *         0 if none of b1 and b2 depends on the other
	 */
	public int compare(Block b1, Block b2) {
		if (isBefore(b1, b2)) {
			// If b1 is before b2 in the dependency graph, b1 is lower than b2
			return -1;
		} else if (isBefore(b2, b1)) {
			// If b2 is before b1 in the dependency graph, b1 is greter than b2
			return 1;
		} else {
			// Else b1 and b2 are "equal" from the dependency point of view
			return 0;
		}
	}

	/**
	 * Tells if b1 is before b2 in the graph of dependencies.
	 */
	public static boolean isBefore(Block b1, Block b2) {
		Platform.getLog().finest("isBefore(" + b1 + ", " + b2 + ")");
		return isBefore(b1, b2, new HashSet<Block>());
	}
	
	private static boolean isBefore(Block b1, Block b2, HashSet<Block> seenBlocks) {
		// b1 is before b2 if we can reach b2 starting from an output pin of b1
		for (Pin p : b1.getPinsOut()) {
			if (!isPinInstantaneouslyDependent(p)) {
				// If this pin does not depend instantaneously 
				// on the inputs of its block, the relations which start
				// from it are not arcs in the dependency graph, 
				// so we skip this pin.
				continue;
			}
			for (Relation r : p.getOutgoingRelations()) {
				Pin input = r.getTarget();
				if (!isPinInstantaneouslyDependent(input)) {
					// If the behavior of the block does not depend
					// instantaneously on the data on this pin, do
					// not consider this relation as an arc in the 
					// graph of dependencies
					continue;
				}
				Block target = input.getInputOf();
				if (target == b2) {
					// We reached b2, so b1 is before b2
					return true;
				}
				if (seenBlocks.contains(target)) {
					// If we already explored the graph from this block, skip it
					continue;
				}
				// Remember we have seen this block
				seenBlocks.add(target);
				// We know that b1 is before target, so b1 is before b2 if target is before b2
				if (isBefore(target, b2, seenBlocks)) {
					return true;
				}
			}
		}
		// If we could not reach b2 from b1, b1 is not before b2
		return false;
	}
	
	public static Map<Block, Integer> assignDependencyRank(CompositeBlock structure) {
		HashMap<Block, Integer> ranks = new HashMap<Block, Integer>();
		for (Block b : structure.getBlocks()) {
			int rank = 0;
			for (Block other : structure.getBlocks()) {
				if (b == other) {
					continue;
				}
				if (isBefore(other, b)) {
					rank++;
				}
			}
			ranks.put(b, rank);
		}
		return ranks;
	}
	
	public static List<List<Pin>> findDependencyLoops(CompositeBlock structure) {
		List<List<Pin>> loops = new LinkedList<List<Pin>>();
		for (Block b : structure.getBlocks()) {
			List<Pin> path = new LinkedList<Pin>();
			findLoopsFrom(b, b, loops, path);
		}
		return loops;
	}
	
	private static void findLoopsFrom(Block start, Block b, List<List<Pin>> loops, List<Pin> path) {
		for (Pin p : b.getPinsOut()) {
			if (isPinInstantaneouslyDependent(p)) {
				path.add(p);
				for (Relation r : p.getOutgoingRelations()) {
					Pin to = r.getTarget();
					if (isPinInstantaneouslyDependent(to)) {
						path.add(to);
						Block owner = to.getInputOf();
						// If the target pin of the relation is not an input, we have
						// reached the border of the structure (the pin is an output
						// pin of the englobing composite), so don't explore further
						if (owner != null) {
							if (owner == start) {
								// We find the starting block, so we have a loop
								loops.add(path);
								// Do not modifiy the path of this loop while searching for other loops
								path = new LinkedList<Pin>(path);
							} else {
								findLoopsFrom(start, owner, loops, path);
							}
						}
						path.remove(path.size() - 1);
					}
				}
				path.remove(path.size() - 1);
			}
		}
	}
	
	public static Block getOwner(Pin p) {
		Block b = p.getInputOf();
		if (b == null) {
			b = p.getOutputOf();
		}
		return b;
	}
	/**
	 * Used to mark a pin as "not instantaneously dependent" (with depends = false).
	 * For an output pin, this means that its value does not depend
	 * on the current inputs of the blocks (state determined output).
	 * For an input pin, this means that the current outputs of the 
	 * block do not depend on the value of this pin (it is only used
	 * to determine the next state of the block).
	 */
	public static void setPinInstantaneousDependency(Pin pin, boolean depends) {
		pin.setValueOfParameter(DE_NO_INSTANTANEOUS_DEPENDENCY, !depends);
	}

	/**
	 * Tells if a pin may depend instantaneously on the inputs of its
	 * block (is not marked as "not instantaneously dependent").
	 */
	public static boolean isPinInstantaneouslyDependent(Pin pin) {
		Object nodep = pin.getValueOfParameter(DE_NO_INSTANTANEOUS_DEPENDENCY);
		if (nodep == null) {
			return true;
		} else {
			return !(Boolean)nodep;
		}
	}
}
