package fr.supelec.modhelx.util;

import java.util.List;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;

import fr.supelec.modhelx.core.Block;
import fr.supelec.modhelx.core.CompositeBlock;
import fr.supelec.modhelx.core.CoreFactory;
import fr.supelec.modhelx.core.InterfaceBlock;
import fr.supelec.modhelx.core.ModelOfComputation;
import fr.supelec.modhelx.core.Pin;
import fr.supelec.modhelx.core.Relation;
import fr.supelec.modhelx.core.Token;
import fr.supelec.modhelx.mocs.delib.DeScenarioEvent;
import fr.supelec.modhelx.mocs.delib.DeScenarioPlayer;
import fr.supelec.modhelx.mocs.delib.DelibFactory;
import fr.supelec.modhelx.mocs.fsm.FsmModelOfComputation;
import fr.supelec.modhelx.mocs.sdf.SdfModelOfComputation;
import fr.supelec.modhelx.mocs.tfsm.TFsmModelOfComputation;

/**
 * This class provides various method that ease the creation of models.
 * 
 * @author Christophe Jacquet
 * @date 2010-04-09
 *
 */
public final class ModelBuilder {
	public static Relation addRelation(CompositeBlock structure, Pin sourcePin, Pin targetPin) {
		Relation r = CoreFactory.eINSTANCE.createRelation();
		r.setSource(sourcePin);
		r.setTarget(targetPin);
		structure.getRelations().add(r);
		return r;
	}

	public static Relation addRelation(InterfaceBlock structure, Pin sourcePin, Pin targetPin) {
		Relation r = CoreFactory.eINSTANCE.createRelation();
		r.setSource(sourcePin);
		r.setTarget(targetPin);
		structure.getRelations().add(r);
		return r;
	}
	
	public static Relation addRelation(InterfaceBlock ib, Pin sourcePin, String targetPinName) {
		return addRelation(ib, sourcePin, pinByName(ib.getInternalModel().getStructure(), targetPinName));
	}
	
	public static Relation addRelation(InterfaceBlock ib, String sourcePinName, Pin targetPin) {
		return addRelation(ib, pinByName(ib.getInternalModel().getStructure(), sourcePinName), targetPin);
	}
	
	/*
	public static void addRelationsFrom(InterfaceBlock ib, Pin sourcePin, String ... pinNames) {
		for(String pinName : pinNames) {
			addRelation(ib, sourcePin, pinByName(ib.getInternalModel().getStructure(), pinName));
		}
	}
	
	public static void addRelationsTo(InterfaceBlock ib, String[] pinNames, Pin targetPin) {
		for(String pinName : pinNames) {
			addRelation(ib, pinByName(ib.getInternalModel().getStructure(), pinName), targetPin);
		}
	}
	*/
	
	private static Pin addPin(List<Pin> pins, String pinName) {
		Pin p = CoreFactory.eINSTANCE.createPin();
		p.setName(pinName);
		pins.add(p);
		
		return p;
	}
	
	public static Pin addPinIn(Block block, String pinName) {
		return addPin(block.getPinsIn(), pinName);
	}
	
	public static Pin addPinOut(Block block, String pinName) {
		return addPin(block.getPinsOut(), pinName);
	}
	
	public static Pin addSDFPinIn(Block block, String pinName, SdfModelOfComputation sdfMoc, int rate) {
		Pin p = addPinIn(block, pinName);
		sdfMoc.setPinRate(p, rate);
		
		return p;
	}

	public static Pin addSDFPinOut(Block block, String pinName, SdfModelOfComputation sdfMoc, int rate) {
		Pin p = addPinOut(block, pinName);
		sdfMoc.setPinRate(p, rate);
		
		return p;
	}

	
	/**
	 * Returns the bin of a block with the specified name.
	 * 
	 * @param block a block
	 * @param pinName the name of the pin
	 * @throws RuntimeException if there is no pin with the specified name
	 * @return the corresponding pin
	 */
	public static Pin pinByName(Block block, String pinName) {
		for(Pin p : block.getPinsIn()) {
			if(pinName.equals(p.getName())) return p;
		}
		
		for(Pin p : block.getPinsOut()) {
			if(pinName.equals(p.getName())) return p;
		}
		
		throw new RuntimeException("Pin '" + pinName + "' not found on block " + block);
	}
	
	public static Pin addPinIfNecessary(Block block, String pinName) {
		Pin pOut = null;
		for(Pin p : block.getPinsOut()) {
			if(p.getName().equals(pinName)) {
				pOut = p;
				break;
			}
		}
		
		if(pOut == null) {
			pOut = CoreFactory.eINSTANCE.createPin();
			pOut.setName(pinName);
			block.getPinsOut().add(pOut);
		}
		
		return pOut;
	}
	
	public static DeScenarioEvent addDeScenarioEvent(DeScenarioPlayer player, int time, String pinName, Object value) {
		DeScenarioEvent evt = DelibFactory.eINSTANCE.createDeScenarioEvent();
		evt.setTime(time);
		evt.setValue(value);

		Pin pOut = addPinIfNecessary(player, pinName);
		evt.setPin(pOut);
		
		player.getEvents().add(evt);
		
		return evt;
	}
	
	/*
	public static DeKeyEvent addDeKeyEvent(DeKeypad keypad, String keyLabel, String pinName, Object value) {
		DeKeyEvent evt = DelibFactory.eINSTANCE.createDeKeyEvent();
		evt.setKey(keyLabel);
		evt.setValue(value);
		
		Pin pOut = addPinIfNecessary(keypad, pinName);
		evt.setPin(pOut);
		
		keypad.getKeys().add(evt);
		return evt;
	}
	*/
	
	public static Token createFSMEvent(ModelOfComputation moc) {
		if(moc instanceof TFsmModelOfComputation) {
			return ((TFsmModelOfComputation)moc).createEvent();
		} else if(moc instanceof FsmModelOfComputation) {
			return ((FsmModelOfComputation)moc).createEvent();
		} else throw new Error("MoC should be (T)FSM, is " + moc.getClass().getSimpleName());
	}
	
	public static EList<Object> makeEList(Object ... elts) {
		EList<Object> l = new BasicEList<Object>();
		for(Object o : elts) l.add(o);
		return l;
	}
}
