package fr.supelec.modhelx.util;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;

/**
 * A base class for storing information which can be backed up and restored.
 * Since serialization is more common than clonability, this class uses 
 * serialization to store a copy of an object in a byte array, and to 
 * restore an object from this byte array.
 * @author Frédéric Boulanger
 *
 * @param <T> The type of the object to backup and restore
 * 
 * Example:
 * class MyState extends StateWithBackup<MyState> {
 *   public int someData;
 * }
 * class MyObject {
 *   private MyState state_ = null;
 *   
 *   public void doSomething() {
 *     state_ = new MyState();
 *     state_.someData = 2;
 *     state_.backup();
 *     state_.someData = 10;
 *     state_ = state_.restore();
 *     // someData is now 2
 *   }
 */
public abstract class StateWithBackup<T extends StateWithBackup<T>> implements Serializable {
	private static final long serialVersionUID = 1L;
	
	/** Serialized backup of this instance */
	private byte backup_[] = null;
	
	/** Serialize this object in the backup_ byte array */
	@SuppressWarnings("unchecked")
	public T backup() {
		try {
			ByteArrayOutputStream stream = new ByteArrayOutputStream();
			new ObjectOutputStream(stream).writeObject(this);
			backup_ = stream.toByteArray();
		} catch (IOException e) {
			backup_ = null;
			e.printStackTrace();
			throw new Error(getClass().getName()+".backup(): IO error while making backup");
		}
		return (T) this;
	}
	
	/** Restore a copy from the backup_ byte array */
	@SuppressWarnings("unchecked")
	public T restore() {
		if (backup_ == null) {
			throw new Error(getClass().getName()+".restore(): attempt to restore from no backup");
		}
		ByteArrayInputStream stream = new ByteArrayInputStream(backup_);
		T read = null;
		try {
			read = (T) new ObjectInputStream(stream).readObject();
		} catch (Exception e) {
			throw new Error(getClass().getName()+".restore(): attempt to restore from no backup");
		}
		return read;
	}
	
	private void writeObject(ObjectOutputStream s) {
		// Do nothing. Do not store the byte array in the backup !
	}
	private void readObject(ObjectInputStream s) {
		// Do nothing. Do not restore the byte array from the backup !
	}
}
