package fr.supelec.modhelx.util;

import static fr.supelec.modhelx.util.ModelBuilder.addPinIn;
import static fr.supelec.modhelx.util.ModelBuilder.addPinOut;
import static fr.supelec.modhelx.util.ModelBuilder.addRelation;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.text.ParseException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import fr.supelec.modhelx.core.AtomicBlock;
import fr.supelec.modhelx.core.Block;
import fr.supelec.modhelx.core.CompositeBlock;
import fr.supelec.modhelx.core.CoreFactory;
import fr.supelec.modhelx.core.Model;
import fr.supelec.modhelx.core.Pin;
import fr.supelec.modhelx.core.time.Clock;
import fr.supelec.modhelx.core.time.TimeFactory;
import fr.supelec.modhelx.mocs.fsmlib.FsmlibFactory;
import fr.supelec.modhelx.mocs.tfsm.TFsmModelOfComputation;
import fr.supelec.modhelx.mocs.tfsm.TfsmFactory;

public class TFSMBuilder {
	private final TFsmModelOfComputation moc;
	private final Model model;
	private final CompositeBlock structure;
	private final Clock clock;
	private final Map<String, Pin> tfsmPins = new HashMap<String, Pin>();
	private final Map<String, AtomicBlock> tfsmStates = new HashMap<String, AtomicBlock>();
	private final Set<String> usedNames = new HashSet<String>();
	
	
	private TFSMBuilder(String modelName) {
		moc = TfsmFactory.eINSTANCE.createTFsmModelOfComputation();
		moc.setName("TFSM");
		
		clock = TimeFactory.eINSTANCE.createLogicalClock();
		moc.setClock(clock);
		
		System.out.println("TFSM Builder clock = " + clock);
		
		structure = CoreFactory.eINSTANCE.createCompositeBlock();
		structure.setName(modelName + "_struct");
		
		// the following parameter is added as a hint for the GUI to identify
		// a composite block that contains a TFSM
		structure.setValueOfParameter("TFSMRole", "TFSMStructure");

		model = CoreFactory.eINSTANCE.createModel();
		model.setName(modelName);
		model.setMoc(moc);
		model.setStructure(structure);
	}
	
	public TFSMBuilder(String modelName, Reader reader) {
		this(modelName);
		
		BufferedReader br = new BufferedReader(reader);
		
		String line;
		int lineNumber = 0;
		do {
			lineNumber++;
			try {
				line = br.readLine();
			} catch(IOException e) {
				line = null;
			}
			if(line != null) {
				try {
					addLine(line);
				} catch(ParseException e) {
					throw new RuntimeException(lineNumber + ":" + e.getErrorOffset() + ": " + e.getMessage(), e);
				}
			}
		} while(line != null);
	}
	
	private void addLine(String line) throws ParseException {
		String[] comp = line.split("\\s+");
		if(comp.length < 1) throw new ParseException("Missing line type.", 0);
		String type = comp[0];
		
		if("State".equals(type) || "InitialState".equals(type)) {
			if(comp.length != 2) throw new ParseException("State lines must have 2 components.", 1);
			String name = comp[1];
			
			if(tfsmStates.get(name) != null) throw new ParseException("State " + name + " already defined.", 2);
			
			AtomicBlock state = CoreFactory.eINSTANCE.createAtomicBlock();
			state.setName(name);
			structure.getBlocks().add(state);
			
			// define it as a state
			moc.setStateRole(state);
			if("InitialState".equals(type)) moc.setInitialStateFeature(state);
			
			// one input, one output pin, used later to connect transitions
			addPinIn(state, "from_trans");
			addPinOut(state, "to_trans");
			
			tfsmStates.put(name, state);
			
		} else if("In".equals(type) || "Out".equals(type)) {
			if(comp.length != 2) throw new ParseException("In/Out lines must have 2 components.", 1);
			String name = comp[1];
			
			if(tfsmPins.get("name") != null) throw new ParseException("Pin " + name + " already defined.", 2);
			Pin thePin = CoreFactory.eINSTANCE.createPin();
			thePin.setName(name);
			tfsmPins.put(name, thePin);
			
			List<Pin> pinList = ("In".equals(type) ? structure.getPinsIn() : structure.getPinsOut());
			pinList.add(thePin);
			
		} else if("Transition".equals(type)) {
			if(comp.length < 4) throw new ParseException("Transition lines must have at least 4 components.", 1);
			String from = comp[1];
			String to = comp[2];
			String guardType = comp[3];
			
			Block fromState = tfsmStates.get(from);
			if(fromState == null) throw new ParseException("State " + from + " undefined.", 2);
			
			Block toState = tfsmStates.get(to);
			if(toState == null) throw new ParseException("State " + to + " undefined.", 2);
			
			List<String> events = new LinkedList<String>();
			List<String> actions = new LinkedList<String>();
			int i;
			for(i=4; i<comp.length; i++) {
				if(comp[i].charAt(0) != '/') events.add(comp[i]);
				else break;
			}
			for(; i<comp.length; i++) {
				if(comp[i].charAt(0) == '/') {
					actions.add(comp[i].substring(1));
				} else throw new ParseException("Only actions are allowed after the first event.", i+1);
			}
			
			// create the transition
			CompositeBlock transition = CoreFactory.eINSTANCE.createCompositeBlock();
			moc.setTransitionRole(transition);
			structure.getBlocks().add(transition);
			
			// create the guard
			Block guard = FsmlibFactory.eINSTANCE.createFsmOrCondition();
			moc.setGuardRole(guard);
			guard.setName(from + "_" + to + "_guard");
			
			if("Or".equals(guardType)) {
				transition.setName("if_" + StringUtils.implode(events, "_or_"));
			} else if("Delay".equals(guardType)) {
				if(events.size() != 1) throw new ParseException("Delay must be followed by the duration.", 5);
				
				int duration;
				try {
					duration = Integer.parseInt(events.get(0));
				} catch(NumberFormatException e) {
					throw new ParseException("The duration after Delay must be an int.", 5);
				}
				
				// there are no events for timed transition
				events.clear();
				
				transition.setName("after_" + duration);
				
				moc.setTimedTransitionFeature(transition);
				moc.setWhenDeadline(guard, moc.getClock().createInstantValue(duration));
			} else throw new ParseException("Illegal guard: " + guardType, 4);
			
			while(usedNames.contains(transition.getName()))
				transition.setName(transition.getName() + "_");
			usedNames.add(transition.getName());
			
			transition.getBlocks().add(guard);			
			
			// create input & output "trigger" pins to the guard
			Pin gInTrig = addPinIn(guard, "in_trig");
			Pin gOutTrig = addPinOut(guard, "out_trig");
			
			// create input & output "trigger" pins to the transition
			Pin inTrig = addPinIn(transition, "in_trig");
			Pin outTrig = addPinOut(transition, "out_trig");
			
			// connect transition triggers and guard triggers
			addRelation(transition, inTrig, gInTrig);
			addRelation(transition, gOutTrig, outTrig);
			
			// connect to the incoming state
			addRelation(structure, fromState.getPinsOut().get(0), inTrig);
			
			// connect to the outgoing state
			addRelation(structure, outTrig, toState.getPinsIn().get(0));
			
			// add an input pin for each event
			for(String event : events) {
				// pin on the guard
				Pin pEventInGuard = addPinIn(guard, event + "_event");
				
				// pin on the transition composite
				Pin pEventInTrans = addPinIn(transition, event + "_event");
				
				// relation between pin on the transition and on the guard
				addRelation(transition, pEventInTrans, pEventInGuard);
				
				// relation between the TFSM input and the transition pin
				Pin outerPin = tfsmPins.get(event);
				if(outerPin == null) throw new ParseException("No such event: " + outerPin, -1);
				addRelation(structure, outerPin, pEventInTrans);
			}
			
			// add action pins and blocks
			for(String action : actions) {
				// pin on the guard
				Pin pActionOutGuard = addPinOut(guard, action + "_action");
				
				// action block
				Block actionBlock = FsmlibFactory.eINSTANCE.createFsmAndCondition();  // TODO: why AndCondition?
				actionBlock.setName(action + "_actblk");
				moc.setActionRole(actionBlock);
				transition.getBlocks().add(actionBlock);
				
				Pin pActionBlkIn = addPinIn(actionBlock, "act_in");
				Pin pActionBlkOut = addPinOut(actionBlock, "act_out");
				
				// connection from guard to action
				addRelation(transition, pActionOutGuard, pActionBlkIn);
				
				// connection from action to transition out pin
				Pin pOut = addPinOut(transition, action);
				
				addRelation(transition, pActionBlkOut, pOut);
				
				// connection from transition out to TFSM out
				Pin pOuter = tfsmPins.get(action);
				if(pOuter == null) throw new ParseException("No such action: " + action, -1);
				addRelation(structure, pOut, pOuter);
			}

		} else throw new ParseException("No such line type: " + type, 0);
	}
	
	public Model getModel() {
		return model;
	}
	

}
