package fr.supelec.modhelx.util.plugins;

import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JPanel;

import fr.supelec.modhelx.core.ModelOfComputation;
import fr.supelec.modhelx.core.Pin;
import fr.supelec.modhelx.core.time.InstantValue;
import fr.supelec.modhelx.util.Numeric;
import fr.supelec.modhelx.util.SdfPluggableBlock;

/**
 * Plug-in for the SdfPluginBlock
 * This plug-in draws a graphical representation of the data received
 * on its inputs. It draws as many plots as it has inputs.
 * 
 * For each input of name <name>, 3 parameters are used to render the plot:
 *   <name>_plotColor is a string which names a static field 
 *                    of class java.awt.Color (e.g. BLACK, RED or BLUE)
 *   <name>_dotSize is an integer which gives the size of the dot 
 *                  drawn for each data sample (0 means no dot)
 *   <name>_lineWidth is an integer which gives the width of the 
 *                    line drawn between data samples (0 means no line)
 * 
 * The following parameters are used globally to plot the data:
 * 	 zoom_x gives the zoom factor on the horizontal scale
 *   zoom_y gives the zoom factor on the vertical scale
 *	 x_origin gives the x value that will correspond to the left side of the plot
 *	 y_origin gives the y value that will correspond to the bottom of the plot
 *	 history gives the number of data samples that are kept for plotting. 
 *           If null or negative, all data samples are kept, else the plot 
 *           will "scroll to the right" as old data is removed from the history
 *           
 * Clicking in the plot moves the plot.
 * Clicking in the plot with the shift key down selects a rectangle to which the plot will zoom.
 *
 * @author Frédéric Boulanger
 *
 */
public class SdfSwingPlotter extends SdfPluggableBlock implements ActionListener {
	private static class PlotParams {
		public Color color = Color.BLACK;
		public int dotsize = 2;
		public int linewidth = 1;
	}
	private static class DataPlotter extends JComponent implements MouseListener, MouseMotionListener {
		private static final long serialVersionUID = 1L;
		private List<Number>[] data_ = null;
		private PlotParams[] params_ = null;
		private int historysize_ = 0;
		private double scale_x_ = 1;
		private int offset_x_ = 0;
		private double scale_y_ = 5;
		private int offset_y_ = 0;
		private Dimension prefsize_ = new Dimension(600, 400);
		private int start_x, start_y, end_x, end_y;
		private boolean zoom_mode_;
		private boolean show_coords_;
		
		private double default_x_scale = 1;
		private double default_y_scale = 1;
		private int default_x_offset = 0;
		private int default_y_offset = 0;
		
		@SuppressWarnings("unchecked")
		public DataPlotter(int width, int histsize) {
			historysize_ = histsize;
			data_ = new List[width];
			for (int i = 0; i < width; i++) {
				data_[i] = new LinkedList<Number>();
			}
			params_ = new PlotParams[width];
			for (int i = 0; i < width; i++) {
				params_[i] = new PlotParams();
			}
			addMouseListener(this);
			addMouseMotionListener(this);
		}
		
		public void setPlotParams(int channel, PlotParams p) {
			params_[channel] = p;
		}
		
		public PlotParams getPlotParams(int channel) {
			return params_[channel];
		}
		
		private double xToValue(int x) {
			return (double)(x - offset_x_) / scale_x_;
		}
		
		private double yToValue(int y) {
			return (double)(offset_y_ + getHeight() - y) / scale_y_;
		}
		
		public void setDefaultView() {
			default_x_offset = offset_x_;
			default_y_offset = offset_y_;
			default_x_scale = scale_x_;
			default_y_scale = scale_y_;
		}
		
		public void restoreDefaultView() {
			offset_x_ = default_x_offset;
			offset_y_ = default_y_offset;
			scale_x_ = default_x_scale;
			scale_y_ = default_y_scale;
		}
		
		@Override
		protected void paintComponent(Graphics g) {
			super.paintComponent(g);
			Graphics2D g2 = (Graphics2D) g.create();
			g2.setColor(Color.WHITE);
			g2.fillRect(0, 0, getWidth(), getHeight());
			int h = getHeight();
			for (int i = 0; i < data_.length; i++) {
				int lastx = 0;
				int lasty = 0;
				for (int j = 0; j < data_[i].size(); j++) {
					int x = offset_x_ + (int) (j * scale_x_);
					int y = offset_y_ + h - (int) (data_[i].get(j).doubleValue() * scale_y_);
					g2 = (Graphics2D) g.create();
					g2.setColor(params_[i].color);
					int dotsize = params_[i].dotsize;
					int linewidth = params_[i].linewidth;
					if (dotsize > 0) {
						g2.fillOval(x-dotsize, y-dotsize, 2*dotsize, 2*dotsize);
					}
					if ((j > 0) && (linewidth > 0)) {
						g2.setStroke(new BasicStroke(linewidth));
						g2.drawLine(lastx, lasty, x, y);
					}
					lastx = x;
					lasty = y;
				}
			}
			if (zoom_mode_) {
				float dashpattern[] = {5f, 5f};
				g2.setColor(Color.DARK_GRAY);
				g2.setStroke(new BasicStroke(1f, BasicStroke.CAP_SQUARE, BasicStroke.JOIN_MITER, 1f, dashpattern, 0f));
				int xmin = Math.min(start_x, end_x);
				int ymin = Math.min(start_y, end_y);
				int xmax = Math.max(start_x, end_x);
				int ymax = Math.max(start_y, end_y);
				g2.drawRect(xmin, ymin, xmax - xmin, ymax - ymin);
				g2.drawString("(" + xToValue(xmin) + "," + yToValue(ymin) + ") -- (" + xToValue(xmax) + "," + yToValue(ymax) + ")", 5, 15);
			} else if (show_coords_) {
				g2.setColor(Color.DARK_GRAY);
				g2.drawString("(" + xToValue(end_x) + "," + yToValue(end_y) + ")", 5, 15);
				
			}
		}
		
		public void setXmin(double xmin) {
			offset_x_ = (int) (- xmin * scale_x_);
		}
		public void setXmax(double xmax) {
			prefsize_.width = offset_x_ + (int) (xmax * scale_x_);
		}
		public void setYmin(double ymin) {
			offset_y_ = (int) (ymin * scale_y_);
		}
		public void setYmax(double ymax) {
			prefsize_.height = (int) (ymax * scale_y_) - offset_y_;
		}
		@Override
		public Dimension getPreferredSize() {
			return prefsize_;
		}
		
		public void setScaleX(double s) {
			scale_x_ = s;
		}
		public double getScaleX() {
			return scale_x_;
		}
		public void setScaleY(double s) {
			scale_y_ = s;
		}
		public double getScaleY() {
			return scale_y_;
		}

		public void setHistoryLength(int length) {
			historysize_ = length;
			for (int i = 0; i < data_.length; i++) {
				while (data_[i].size() > historysize_) {
					data_[i].remove(0);
				}
			}
		}
		public void addData(int channel, Number data) {
			data_[channel].add(data);
			if ((historysize_ > 0) && (data_[channel].size() > historysize_)) {
				data_[channel].remove(0);
			}
		}

		public void mouseClicked(MouseEvent e) {			
		}

		public void mouseEntered(MouseEvent e) {			
		}

		public void mouseExited(MouseEvent e) {
			show_coords_ = false;
			repaint();
		}

		public void mousePressed(MouseEvent e) {
			start_x = e.getX();
			start_y = e.getY();
			zoom_mode_ = (e.getModifiers() & InputEvent.SHIFT_MASK) != 0;
		}

		public void mouseReleased(MouseEvent e) {
			end_x = e.getX();
			end_y = e.getY();
			if (zoom_mode_) {
				doZoomToRectangle(start_x, start_y, end_x, end_y);
			}
			zoom_mode_ = false;
		}

		public void mouseDragged(MouseEvent e) {
			end_x = e.getX();
			end_y = e.getY();
			if (!zoom_mode_) {
				doPanToRectangle(start_x, start_y, end_x, end_y);
				start_x = end_x;
				start_y = end_y;
			}
			repaint();
		}

		public void mouseMoved(MouseEvent e) {
			end_x = e.getX();
			end_y = e.getY();
			show_coords_ = true;
			repaint();
		}
		
		private void doZoomToRectangle(int x1, int y1, int x2, int y2) {
			int xmin = Math.min(x1,x2);
			int xmax = Math.max(x1,x2);
			int ymin = Math.min(y1,y2);
			int ymax = Math.max(y1,y2);
			double hmin = xToValue(xmin);
			double vmin = yToValue(ymax);
			scale_x_ = scale_x_ * ((double)getWidth()) / (xmax - xmin);
			scale_y_ = scale_y_ * ((double)getHeight()) / (ymax - ymin);
			setXmin(hmin);
			setYmin(vmin);
			repaint();
		}
		
		private void doPanToRectangle(int x1, int y1, int x2, int y2) {
			int dx = x2 - x1;
			int dy = y2 - y1;
			offset_x_ += dx;
			offset_y_ += dy;
			repaint();
		}
	}
	
	private static final String parameterNames_[] = {
		"zoom_x",
		"zoom_y",
		"x_origin",
		"y_origin",
		"history"
	};

	private JFrame frame_ = null;
	private DataPlotter plotter_ = null;
	private Number data_[] = null;
	
	@Override
	public void setup(ModelOfComputation moc, InstantValue initialTime) {
		super.setup(moc, initialTime);
		int size = getPinsIn().size();
		plotter_ = new DataPlotter(size, 0);
		data_ = new Number[size];
		for (Pin p : getPinsIn()) {
			if (getMoC().getPinRate(p) != 1) {
				throw new Error(getClass().getName()+".setup(): ("+p.getName()+") I support only inputs with rate 1");
			}
		}
		frame_ = new JFrame(getName());
		JPanel controls = new JPanel();
		JButton zout = new JButton("Zoom out");
		zout.setActionCommand("ZOOMOUT");
		zout.addActionListener(this);
		controls.add(zout);
		JButton zin = new JButton("Zoom in");
		zin.setActionCommand("ZOOMIN");
		zin.addActionListener(this);
		controls.add(zin);
		JButton zdef = new JButton("Default view");
		zdef.setActionCommand("ZOOMDEF");
		zdef.addActionListener(this);
		controls.add(zdef);
		JPanel data = new JPanel(new BorderLayout());
		data.add(plotter_, BorderLayout.CENTER);
		for (int ch = 0; ch < getPinsIn().size(); ch++) {
			PlotParams pp = plotter_.getPlotParams(ch);
			Pin p = getPinsIn().get(ch);
			// color
			Object param = getValueOfParameter(p.getName() + "_plotColor");
			if ((param != null) && (param instanceof String)) {
				Object c = null;
				try {
					c = Color.class.getDeclaredField((String) param).get(null);
				} catch (Exception e) {
					System.err.println("# \"" + param + "\" is not a valid color name");
					c = null;
				}
				if ((c != null) && (c instanceof Color)) {
					pp.color = (Color) c;
				}
			}
			// dot size
			param = getValueOfParameter(p.getName() + "_dotSize");
			if (param != null) {
				pp.dotsize = Numeric.intValueOfObject(param);
			}
			// line width
			param = getValueOfParameter(p.getName() + "_lineWidth");
			if (param != null) {
				pp.linewidth = Numeric.intValueOfObject(param);
			}
			plotter_.setPlotParams(ch, pp);
		}
		Object param = getValueOfParameter("zoom_x");
		if (param != null) {
			plotter_.setScaleX(Numeric.doubleValueOfObject(param));
		}
		param = getValueOfParameter("zoom_y");
		if (param != null) {
			plotter_.setScaleY(Numeric.doubleValueOfObject(param));
		}
		param = getValueOfParameter("x_min");
		if (param != null) {
			plotter_.setXmin(Numeric.doubleValueOfObject(param));
		}
		param = getValueOfParameter("x_max");
		if (param != null) {
			plotter_.setXmax(Numeric.doubleValueOfObject(param));
		}
		param = getValueOfParameter("y_min");
		if (param != null) {
			plotter_.setYmin(Numeric.doubleValueOfObject(param));
		}
		param = getValueOfParameter("y_max");
		if (param != null) {
			plotter_.setYmax(Numeric.doubleValueOfObject(param));
		}
		plotter_.setDefaultView();
		param = getValueOfParameter("history");
		if (param != null) {
			plotter_.setHistoryLength(Numeric.intValueOfObject(param));
		}
		frame_.getContentPane().setLayout(new BorderLayout());
		frame_.getContentPane().add(controls, BorderLayout.NORTH);
		frame_.getContentPane().add(data, BorderLayout.CENTER);
		frame_.pack();
		frame_.setVisible(true);
	}
	
	@Override
	public void wrapup(ModelOfComputation moc) {
		super.wrapup(moc);
		frame_.dispose();
	}
	
	@Override
	public void reset(ModelOfComputation moc, InstantValue time) {
		super.reset(moc, time);
		for (int i = 0; i < data_.length; i++) {
			data_[i] = null;
		}
	}
	
	@Override
	public void endOfSnapshot(ModelOfComputation moc) {
		super.endOfSnapshot(moc);
		for (int i = 0; i < data_.length; i++) {
			plotter_.addData(i, data_[i]);
		}
		plotter_.repaint();
	}
	
	@Override
	public void update(ModelOfComputation moc) {
		super.update(moc);
		for (int i = 0; i < data_.length; i++) {
			data_[i] = Numeric.doubleValueOfObject(
					getMoC().getValueOfToken(
							getPinsIn().get(i).getToken(true)
					)
			);
		}
	}

	@Override
	public List<String> parameterNames() {
		return Arrays.asList(parameterNames_);
	}

	public void actionPerformed(ActionEvent e) {
		String cmd = e.getActionCommand();
		if (cmd.equals("ZOOMOUT")) {
			plotter_.setScaleX(plotter_.getScaleX() / 2);
			plotter_.setScaleY(plotter_.getScaleY() / 2);
		} else if (cmd.equals("ZOOMIN")) {
			plotter_.setScaleX(plotter_.getScaleX() * 2);
			plotter_.setScaleY(plotter_.getScaleY() * 2);
		} else if (cmd.equals("ZOOMDEF")) {
			plotter_.restoreDefaultView();
		}
		plotter_.repaint();
	}
}
