package fr.supelec.modhelx.util.plugins;

import java.awt.Color;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;

import javax.swing.JFrame;
import javax.swing.JScrollBar;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.SwingUtilities;
import javax.swing.table.DefaultTableModel;

import fr.supelec.modhelx.core.ModelOfComputation;
import fr.supelec.modhelx.core.Pin;
import fr.supelec.modhelx.core.time.InstantValue;
import fr.supelec.modhelx.util.SdfPluggableBlock;

/**
 * A block which displays its inputs in a JTable inside a JFrame
 * The table has a first column in which the time of the snapshot is shown.
 * Subsequent columns display the values read on the inputs. Each row
 * displays a sample read from an input. If the inputs do not have the
 * same data consumption rate, blank cells are left in the column of
 * the less consuming inputs so that data read during the same snapshot
 * are kept horizontally aligned.
 * 
 * @author Frédéric Boulanger
 *
 */
public class SdfSwingPrinter extends SdfPluggableBlock {
	private static final String parameterNames_[] = {
	};
	
	private JFrame frame_ = null;
	private JTable table_ = null;
	private JScrollBar vscroll_ = null;
	private Object[][] values_ = null;
	private int depth_ = 1;
	private int nextrow_ = 0;
	
	@Override
	public void setup(ModelOfComputation moc, InstantValue initialTime) {
		super.setup(moc, initialTime);
		int size = getPinsIn().size();
		values_ = new Object[size][];
		depth_ = 1;
		for (int i = 0; i < size; i++) {
			int rate = getMoC().getPinRate(getPinsIn().get(i));
			depth_ = Math.max(rate, depth_);
			values_[i] = new Object[rate];
		}
		frame_ = new JFrame(getName());
		List<String> names = new LinkedList<String>();
		names.add("Time"); // room for "time" column
		for (Pin p : getPinsIn()) {
			names.add(p.getName());
		}
		table_ = new JTable();
		table_.setModel(new DefaultTableModel(names.toArray(), 0));
		table_.setShowGrid(true);
		table_.setGridColor(Color.LIGHT_GRAY);
		table_.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
		nextrow_ = 0;
		JScrollPane scroll = new JScrollPane(table_,
				                             JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
				                             JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		vscroll_ = scroll.getVerticalScrollBar();
		frame_.getContentPane().add(scroll);
		frame_.pack();
		frame_.setVisible(true);
	}
	
	@Override
	public void wrapup(ModelOfComputation moc) {
		super.wrapup(moc);
		frame_.dispose();
	}
	
	@Override
	public void reset(ModelOfComputation moc, InstantValue time) {
		super.reset(moc, time);
		for (int i = 0; i < values_.length; i++) {
			for (int j = 0; j < values_[i].length; j++) {
				values_[i][j] = null;
			}
		}
	}
	
	@Override
	public void endOfSnapshot(ModelOfComputation moc) {
		super.endOfSnapshot(moc);
		for (int i = 0; i < depth_; i++) {
			((DefaultTableModel)(table_.getModel())).addRow((Vector<Object>)null);
		}
		table_.setValueAt(getMoC().getTimeOfCurrentSnapshot().getDenotedInstant().get(0).getDate(), nextrow_, 0);
		for (int i = 0; i < values_.length; i++) {
			for (int j = 0; j < values_[i].length; j++) {
				table_.setValueAt(values_[i][j], nextrow_ + j, 1 + i);
			}
		}
		nextrow_ += depth_;
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				vscroll_.setValue(vscroll_.getMaximum());
			}
		});
	}
	
	@Override
	public void update(ModelOfComputation moc) {
		super.update(moc);
		for (int i = 0; i < values_.length; i++) {
			for (int j = 0; j < values_[i].length; j++) {
				values_[i][j] = getMoC().getValueOfToken(getPinsIn().get(i).getToken(true));
			}
		}
	}

	@Override
	public List<String> parameterNames() {
		return Arrays.asList(parameterNames_);
	}
}
